@@ -1,24 +1,29 @@
 import ast
-import sys
+from itertools import chain
+import os
 import re
 import string
+import sys
 
 from pathlib import Path
 
-from typing import Callable, List, Optional, Set, TextIO, Tuple, Union
-from deadcode.data_types import Args
+from typing import Callable, Dict, List, Optional, Set, TextIO, Union, Iterable
+from deadcode.constants import UnusedCodeType
+from deadcode.data_types import Args, Part
 from deadcode.visitor.code_item import CodeItem
 from deadcode.visitor.utils import LoggingList, LoggingSet
 from deadcode.visitor import utils
 from deadcode.actions.parse_abstract_syntax_tree import parse_abstract_syntax_tree
 
-# TODO: from deadcode.visitor import noqa
+from deadcode.visitor import noqa
+from deadcode.utils.print_ast import print_ast as show  # noqa: F401
 from deadcode.visitor import lines
 from deadcode.visitor.ignore import (
-    # TODO: ERROR_CODES,
+    ERROR_CODES,
     IGNORED_VARIABLE_NAMES,
     _get_unused_items,
     _match,
+    _match_many,
     _assigns_special_variable__all__,
     _ignore_class,
     _ignore_import,
@@ -31,26 +36,13 @@ from deadcode.visitor.ignore import (
 class DeadCodeVisitor(ast.NodeVisitor):
     """Finds dead code."""
 
-    # TODO: Should work with only one file.
-    # The state should be passed as function argument.
-    # The state should be passed as defined.
-
-    # Should simplify the code by using functional programming.
-
-    # First step is to make the parameters work.
-
     def __init__(self, filenames: List[str], args: Args) -> None:
-        # Dict[Filename, AbstractSyntaxTree]
-        # self.abstract_syntax_trees_of_files = abstract_syntax_trees_of_files
         self.filenames = filenames
 
         self.args = args
 
         self.ignore_decorators: List[str] = []
 
-        # ignore_definitions=[],
-        # ignore_definitions_if_inherits_from=[],
-
         verbose = False
         # verbose = args.verbose
         self.verbose = verbose
@@ -62,19 +54,29 @@ class DeadCodeVisitor(ast.NodeVisitor):
         self.defined_methods: List[CodeItem] = LoggingList("method", verbose)
         self.defined_props: List[CodeItem] = LoggingList("property", verbose)
         self.defined_vars: List[CodeItem] = LoggingList("variable", verbose)
+        self.unused_file: List[CodeItem] = LoggingList("unused_file", verbose)
         self.unreachable_code: List[CodeItem] = LoggingList("unreachable_code", verbose)
 
-        # The purpose and the interface of the visitor is not known
-        # Lets investigate the vulture code and pass here what ever I have.
-
         self.used_names: Set[str] = LoggingSet("name", self.verbose)
 
         self.filename = Path()
         self.code: List[str] = []
-        # self.found_dead_code_or_error = False
 
-        # Workaround
-        # self.noqa_lines = {}
+        # Note: scope is a stack containing current module name, class names, function names
+        self.scope_parts: List[str] = []
+
+        # This flag is used to stop registering code definitions in a code item
+        # during recursive its parsing
+        self.should_ignore_new_definitions = False
+
+        self.noqa_lines: Dict[str, Set[int]] = {}
+
+    @property
+    def scope(self) -> str:
+        return ".".join(self.scope_parts)
+
+    def add_used_name(self, name: str, scope: Optional[str] = None) -> None:
+        self.used_names.add(name)
 
     # def scan(self, code, filename=""):
     #     filename = Path(filename)
@@ -151,12 +153,28 @@ class DeadCodeVisitor(ast.NodeVisitor):
         #     self.scan(module_string, filename=module)
 
         # TODO: should parse the AST here.
-        for filename in self.filenames:
-            with open(filename) as f:
-                node = parse_abstract_syntax_tree(f.read(), args=self.args, filename=filename)
+        for file_path in self.filenames:
+            with open(file_path) as f:
+                filename = os.path.basename(file_path)
+                module_name = os.path.splitext(filename)[0]
+                self.scope_parts = [module_name]
 
-            self.filename = Path(filename)
-            self.visit(node)
+                file_content = f.read()
+                if file_content.strip() or (filename.startswith("__") and filename.endswith("__.py")):
+                    self.noqa_lines = noqa.parse_noqa(file_content)
+
+                    node = parse_abstract_syntax_tree(file_content, args=self.args, filename=file_path)
+                    self.filename = Path(file_path)
+                    self.visit(node)
+                else:
+                    self.unused_file.append(
+                        CodeItem(
+                            name=filename,
+                            type_="unused_file",
+                            filename=Path(file_path),
+                            message="Empty file",
+                        )
+                    )
 
         # unique_imports = {item.name for item in self.defined_imports}
         # for import_name in unique_imports:
@@ -173,30 +191,24 @@ class DeadCodeVisitor(ast.NodeVisitor):
         #         module_string = module_data.decode("utf-8")
         #         self.scan(module_string, filename=path)
 
-    def get_unused_code_items(self, sort_by_size: bool = False) -> List[CodeItem]:
+    def get_unused_code_items(self, sort_by_size: bool = False) -> Iterable[CodeItem]:
         """
         Return ordered list of unused CodeItem objects.
         """
 
-        def by_name(item: CodeItem) -> Tuple[str, int]:
-            return (str(item.filename).lower(), item.first_lineno)
-
-        def by_size(item: CodeItem) -> Tuple[int, str, int]:
-            # TODO: default sorting should be by type or by filename
-            return (item.size,) + by_name(item)
-
-        unused_code = (
-            self.unused_attrs
-            + self.unused_classes
-            + self.unused_funcs
-            + self.unused_imports
-            + self.unused_methods
-            + self.unused_props
-            + self.unused_vars
-            + self.unreachable_code
+        unused_code = chain(
+            self.unused_attrs,
+            self.unused_classes,
+            self.unused_funcs,
+            self.unused_imports,
+            self.unused_methods,
+            self.unused_props,
+            self.unused_vars,
+            self.unreachable_code,
+            self.unused_file,
         )
 
-        return sorted(unused_code, key=by_size if sort_by_size else by_name)
+        return sorted(unused_code, key=lambda item: (item.filename, item.name_line or 0))
 
     # TODO: investigate whitelisting options
     # def report(self, sort_by_size=False, make_whitelist=False):
@@ -212,31 +224,31 @@ class DeadCodeVisitor(ast.NodeVisitor):
     #     return self.found_dead_code_or_error
 
     @property
-    def unused_classes(self) -> List[CodeItem]:
+    def unused_classes(self) -> Iterable[CodeItem]:
         return _get_unused_items(self.defined_classes, self.used_names)
 
     @property
-    def unused_funcs(self) -> List[CodeItem]:
+    def unused_funcs(self) -> Iterable[CodeItem]:
         return _get_unused_items(self.defined_funcs, self.used_names)
 
     @property
-    def unused_imports(self) -> List[CodeItem]:
+    def unused_imports(self) -> Iterable[CodeItem]:
         return _get_unused_items(self.defined_imports, self.used_names)
 
     @property
-    def unused_methods(self) -> List[CodeItem]:
+    def unused_methods(self) -> Iterable[CodeItem]:
         return _get_unused_items(self.defined_methods, self.used_names)
 
     @property
-    def unused_props(self) -> List[CodeItem]:
+    def unused_props(self) -> Iterable[CodeItem]:
         return _get_unused_items(self.defined_props, self.used_names)
 
     @property
-    def unused_vars(self) -> List[CodeItem]:
+    def unused_vars(self) -> Iterable[CodeItem]:
         return _get_unused_items(self.defined_vars, self.used_names)
 
     @property
-    def unused_attrs(self) -> List[CodeItem]:
+    def unused_attrs(self) -> Iterable[CodeItem]:
         return _get_unused_items(self.defined_attrs, self.used_names)
 
     def _log(self, *args, file: Optional[TextIO] = None, force: bool = False) -> None:  # type: ignore
@@ -267,7 +279,7 @@ class DeadCodeVisitor(ast.NodeVisitor):
                 ignore=_ignore_import,
             )
             if alias is not None:
-                self.used_names.add(name_and_alias.name)
+                self.add_used_name(name_and_alias.name)
 
     def _handle_conditional_node(self, node: Union[ast.If, ast.IfExp, ast.While], name: str) -> None:
         if utils.condition_is_always_false(node.test):
@@ -276,7 +288,7 @@ class DeadCodeVisitor(ast.NodeVisitor):
                 name,
                 node,
                 last_node=node.body if isinstance(node, ast.IfExp) else node.body[-1],
-                message=f"unsatisfiable '{name}' condition",
+                message=f"Unsatisfiable `{name}` condition",
             )
         elif utils.condition_is_always_true(node.test):
             else_body = node.orelse
@@ -285,7 +297,7 @@ class DeadCodeVisitor(ast.NodeVisitor):
                     self.unreachable_code,
                     name,
                     else_body,  # type: ignore
-                    message="unreachable 'else' expression",
+                    message="Unreachable `else` expression",
                 )
             elif else_body:
                 self._define(
@@ -293,7 +305,7 @@ class DeadCodeVisitor(ast.NodeVisitor):
                     "else",
                     else_body[0],  # type: ignore
                     last_node=else_body[-1],  # type: ignore
-                    message="unreachable 'else' block",
+                    message="Unreachable `else` block",
                 )
             elif name == "if":
                 # Redundant if-condition without else block.
@@ -301,7 +313,7 @@ class DeadCodeVisitor(ast.NodeVisitor):
                     self.unreachable_code,
                     name,
                     node,
-                    message="redundant if-condition",
+                    message="Redundant `if` condition",
                 )
 
     def _define(
@@ -320,35 +332,33 @@ class DeadCodeVisitor(ast.NodeVisitor):
             f"ignore_definitions_if_inherits_from={self.args.ignore_definitions_if_inherits_from}"
         )
 
-        def ignored(lineno: int) -> bool:
+        def ignored(lineno: int, type_: UnusedCodeType) -> bool:
             return bool(
                 (ignore and ignore(self.filename, name))
                 or _match(name, self.args.ignore_names)
                 or _match(self.filename, self.args.ignore_names_in_files)
-                # TODO: noqa comments should be supported
-                # or noqa.ignore_line(self.noqa_lines, lineno, ERROR_CODES[typ])
+                or self.should_ignore_new_definitions
+                or noqa.ignore_line(self.noqa_lines, lineno, ERROR_CODES[type_])
             )
 
         last_node = last_node or first_node
-        typ: str = collection.typ  # type: ignore
+        type_: UnusedCodeType = collection.type_  # type: ignore
         first_lineno = lines.get_first_line_number(first_node)
 
-        if ignored(first_lineno):
-            self._log(f'Ignoring {typ} "{name}"')
+        if ignored(first_lineno, type_=type_):
+            self._log(f'Ignoring {type_} "{name}"')
         else:
             last_lineno = lines.get_last_line_number(last_node)
 
             collection.append(
                 CodeItem(
                     name=name,
-                    typ=typ,
+                    type_=type_,
                     filename=self.filename,
-                    first_lineno=first_lineno,
-                    last_lineno=last_lineno,
-                    first_column=last_node.col_offset,
-                    last_column=last_node.end_col_offset,
-                    name_line=last_node.lineno,
-                    name_column=last_node.col_offset,
+                    code_parts=[Part(first_lineno, last_lineno, last_node.col_offset, last_node.end_col_offset or 0)],
+                    scope=self.scope,
+                    name_line=last_node.lineno,  # TODO: Maybe this should be a property?
+                    name_column=last_node.col_offset,  # TODO: Maybe this should be a property?
                     message=message,
                 )
             )
@@ -372,7 +382,7 @@ class DeadCodeVisitor(ast.NodeVisitor):
         if isinstance(node.ctx, ast.Store):
             self._define(self.defined_attrs, node.attr, node)
         elif isinstance(node.ctx, ast.Load):
-            self.used_names.add(node.attr)
+            self.add_used_name(node.attr)
 
     def visit_BinOp(self, node: ast.BinOp) -> None:
         """
@@ -391,7 +401,7 @@ class DeadCodeVisitor(ast.NodeVisitor):
         ):
             attr_name_arg = node.args[1]
             if isinstance(attr_name_arg, ast.Str):
-                self.used_names.add(attr_name_arg.s)
+                self.add_used_name(attr_name_arg.s)
 
         # Parse variable names in new format strings:
         # "{my_var}".format(**locals())
@@ -420,7 +430,7 @@ class DeadCodeVisitor(ast.NodeVisitor):
             vars = re.sub(r"\[\w*\]", "", field_name).split(".")
             for var in vars:
                 if is_identifier(var):
-                    self.used_names.add(var)
+                    self.add_used_name(var)
 
     @staticmethod
     def _is_locals_call(node: ast.AST) -> bool:
@@ -447,17 +457,17 @@ class DeadCodeVisitor(ast.NodeVisitor):
         first_arg = node.args.args[0].arg if node.args.args else None
 
         if "@property" in decorator_names:
-            typ = "property"
+            type_ = "property"
         elif "@staticmethod" in decorator_names or "@classmethod" in decorator_names or first_arg == "self":
-            typ = "method"
+            type_ = "method"
         else:
-            typ = "function"
+            type_ = "function"
 
         if any(_match(name, self.ignore_decorators) for name in decorator_names):
-            self._log(f'Ignoring {typ} "{node.name}" (decorator whitelisted)')
-        elif typ == "property":
+            self._log(f'Ignoring {type_} "{node.name}" (decorator whitelisted)')
+        elif type_ == "property":
             self._define(self.defined_props, node.name, node)
-        elif typ == "method":
+        elif type_ == "method":
             self._define(self.defined_methods, node.name, node, ignore=_ignore_method)
         else:
             self._define(self.defined_funcs, node.name, node, ignore=_ignore_function)
@@ -477,7 +487,7 @@ class DeadCodeVisitor(ast.NodeVisitor):
 
     def visit_Name(self, node: ast.Name) -> None:
         if isinstance(node.ctx, (ast.Load, ast.Del)) and node.id not in IGNORED_VARIABLE_NAMES:
-            self.used_names.add(node.id)
+            self.add_used_name(node.id)
         elif isinstance(node.ctx, (ast.Param, ast.Store)):
             self._define_variable(node.id, node)
 
@@ -486,7 +496,7 @@ class DeadCodeVisitor(ast.NodeVisitor):
             assert isinstance(node.value, (ast.List, ast.Tuple))
             for elt in node.value.elts:
                 if isinstance(elt, ast.Str):
-                    self.used_names.add(elt.s)
+                    self.add_used_name(elt.s)
 
     def visit_While(self, node: ast.While) -> None:
         self._handle_conditional_node(node, "while")
@@ -494,9 +504,39 @@ class DeadCodeVisitor(ast.NodeVisitor):
     # def visit_MatchClass(self, node: ast.MatchClass) -> None:  # type: ignore
     def visit_MatchClass(self, node) -> None:  # type: ignore
         for kwd_attr in node.kwd_attrs:
-            self.used_names.add(kwd_attr)
+            self.add_used_name(kwd_attr)
 
     def visit(self, node: ast.AST) -> None:
+        # TODO: use decorator for this code chunk
+        was_scope_increased = True
+        if isinstance(node, ast.ClassDef):
+            self.scope_parts.append(node.name)
+        elif isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
+            self.scope_parts.append(node.name)
+        else:
+            was_scope_increased = False
+
+        # TODO: use decorator for this code chunk
+        should_turn_off_ignore_new_definitions = False
+
+        node_name = getattr(node, "name", None)
+        bases_attr = getattr(node, "bases", [])
+        bases = [base.id for base in bases_attr if getattr(base, "id", None)]
+
+        if (
+            # Name is in ignore_definitions
+            node_name
+            and _match(node_name, self.args.ignore_definitions)
+        ) or (
+            # Class inherits from ignore_definitions_if_inherits_from
+            bases_attr
+            and bases
+            and _match_many(bases, self.args.ignore_definitions_if_inherits_from)
+        ):
+            if not self.should_ignore_new_definitions:
+                self.should_ignore_new_definitions = True
+                should_turn_off_ignore_new_definitions = True
+
         method_name = "visit_" + node.__class__.__name__
         visitor = getattr(self, method_name, None)
         if self.verbose:
@@ -507,6 +547,12 @@ class DeadCodeVisitor(ast.NodeVisitor):
         if visitor:
             visitor(node)
 
+        # Class inherits from ignore_bodies_if_inherits_from
+        if bases and _match_many(bases, self.args.ignore_bodies_if_inherits_from):
+            if not self.should_ignore_new_definitions:
+                self.should_ignore_new_definitions = True
+                should_turn_off_ignore_new_definitions = True
+
         # There isn't a clean subset of node types that might have type
         # comments, so just check all of them.
         type_comment = getattr(node, "type_comment", None)
@@ -516,6 +562,14 @@ class DeadCodeVisitor(ast.NodeVisitor):
 
         self.generic_visit(node)
 
+        # TODO: use decorator for this code chunk
+        if should_turn_off_ignore_new_definitions:
+            self.should_ignore_new_definitions = False
+
+        # TODO: use decorator for this code chunk
+        if was_scope_increased:
+            self.scope_parts.pop()
+
     def _handle_ast_list(self, ast_list: List[ast.AST]) -> None:
         """
         Find unreachable nodes in the given sequence of ast nodes.
